/*
  Name: object.h
  Author: Master Lenman
  Project: SpeedHack 2009
  Date: 10.07.2009
  Description: Base class for all objects
*/

#ifndef HEADER_OBJECT
#define HEADER_OBJECT

#include <string>
#include <allegro.h>

#include "cmap.h"

bool sortCObjects(cobject *a, cobject *b);

extern const int moveDelta[4][2];

class cobject
{
private:
	int posX, posY;
	bool destroyGfx;
	BITMAP *gfx;
	int gfxPosX, gfxPosY;
	int sizeX, sizeY;
	int dir;
	int moveDir;
	int curAnimation;
	int animationCounter;
	int speed;
	int weapon;
	int weaponsAvailible;
	//Size of the collision box
	int collision[4];
	bool walkable;
	bool collectable;
	bool characterAnim;
	bool moving;
	int aiType;
	int aiObstacle;
	int aiRoaming;
	int life, maxLife;
	int spawnType, spawnFrequency;
	int weaponCooldown;
	std::string id, followupId;
	int drawBullet;
	
public:
	cobject();
	virtual ~cobject();
	virtual cobject &operator=(const cobject &temp);
	bool operator<(const cobject &temp);
	virtual void draw(int camX, int camY);
	void drawWeapon(int destX, int destY);
	int tick(cmap &theMap);
	
	virtual void loadTemplate(std::string id, std::string path);
	
	void setPos(int newX, int newY);
	void setPosX(int newX);
	void setPosY(int newY);
	void substractLife(int dLife);
	void setLife(int nLife);
	void setWeaponsAvailible(int nAvai);
	
	int getPosX() const;
	int getPosY() const;
	std::string getId() const;
	std::string getFollowupId() const;
	const int *getCollision() const;
	BITMAP *getGfx() const;
	int getGfxPosX() const;
	int getGfxPosY() const;
	int getSizeX() const;
	int getSizeY() const;
	int getLife() const;
	int getSpeed() const;
	bool getWalkable() const;
	bool getCollectable() const;
	bool getCharacterAnim() const;
	int getAiType() const;
	int getSpawnType() const;
	int getSpawnFrequency() const;
	int getWeapon() const;
};

#endif
