/*
  Name: map.cpp
  Author: Master Lenman
  Project: SpeedHack 2009
  Date: 10.07.2009
  Description: The mapclass for the game
*/

#include <string>
#include <vector>
#include <list>
#include <fstream>
#include <allegro.h>
#include <cmath>

#include "main.h"
#include "cobject.h"
#include "game.h"
#include "cmap.h"
#include "global.h"
#include "errors.h"
#include "menues.h"
#include "keyStatesClass.h"

using namespace std;

cmap::cmap()
{
	sizeX = sizeY = 0;
	tilesetGfx = 0;
	entryGfx = loadBitmap("media/tilesets/entrance.png");
	icons[0] = loadBitmap("media/objects/medikiticon.png");
	icons[1] = loadBitmap("media/objects/ammoicon.png");
	floorX = floorY = floorSizeX = floorSizeY = 0;
	player = 0;
	clockColorCounter = 0;
	fireTemplate = 0;
	leftDropzone = false;
	inventoryItems = 0;
}

cmap::~cmap()
{
	for(int i = 0; i < 4; ++i)
	{
		infos->availibleSaves[i] += availible[i];
	}
	for(int i = 0; i < 2; ++i)
	{
		infos->availibleKills[i] += availible[i + 4];
	}
	destroy_bitmap(entryGfx);
	for(int i = 0; i < 2; ++i)
	{
		destroy_bitmap(icons[i]);
	}
	if(tilesetGfx)
	{
		destroy_bitmap(tilesetGfx);
	}
	for(objListIter ter = objects.begin(); ter != objects.end(); ++ter)
	{
		if((*ter)->getAiType() == 7)
		{
			++totalStats->endFightSurvivors;
		}
		delete (*ter);
	}
}

cobject *cmap::findTemplate(string name)
{
	for(objListIter ter = objectTemplates.begin(); ter != objectTemplates.end(); ++ter)
	{
		if((*ter)->getId() == name)
		{
			return *ter;
		}
	}
	cobject *newTemplate = new cobject();
	newTemplate->loadTemplate(name, "media/objects/" + name + ".txt");
	objectTemplates.push_back(newTemplate);
	return newTemplate;
}

void cmap::load(string path, gameInfos *tInfos, int weaponsUnlocked, bool firefighterUnl, bool lastLevel, gameInfos totals)
{
	firefighterUnlocked = firefighterUnl;
	infos = tInfos;
	totalStats = &totals;
	this->lastLevel = lastLevel;
	
	ifstream mapFile(path.c_str());
	if(!mapFile)
	{
		errors::mapLoadError(path, "Cannot open file!");
	}
	
	mapFile >> sizeX >> sizeY;
	mapFile >> entranceX >> entranceY;
	if(sizeX < SCREEN_W || sizeY < SCREEN_H)
	{
		errors::mapLoadError(path, "Maps is too small! (Minimum: " + i2str(SCREEN_W) + "x" + i2str(SCREEN_H) + ")");
	}
	
	string floorPath;
	mapFile >> floorPath;
	tilesetGfx = loadBitmap(floorPath);
	
	mapFile >> floorX >> floorY >> floorSizeX >> floorSizeY;
	
	//mapFile >> playerX >> playerY;
	
	mapFile >> timeLimit;
	timeLeft = timeLimit * 60;
	for(int i = 0; i < 4; ++i)
	{
		mapFile >> targets[i];
	}
	
	for(int i = 0; i < 6; ++i)
	{
		availible[i] = 0;
	}
	
	string nextObjName;
	mapFile >> nextObjName;
	int refPosX = 0, refPosY = 0;
	while(nextObjName != "endobjects")
	{
		int objX, objY;
		mapFile >> objX >> objY;
		
		if(nextObjName == "refpos")
		{
			refPosX = objX;
			refPosY = objY;
			mapFile >> nextObjName;
			continue;
		}
		
		objX += refPosX;
		objY += refPosY;
		
		cobject *templ = findTemplate(nextObjName);
		cobject *newObj = new cobject();
		*newObj = *templ;
		newObj->setPos(objX, objY);
		objects.push_back(newObj);
		
		if(nextObjName == "player")
		{
			if(!player)
			{
				player = newObj;
				player->setWeaponsAvailible(weaponsUnlocked);
			}
			else
			{
				errors::mapLoadError(path, "Multiple player objects!");
			}
		}
		switch(nextObjName[0])
		{
			//Hostage:
			case 'h':
				++availible[0];
				break;
			case 'f':
				++availible[1];
				break;
			case 'm':
				++availible[2];
				break;
			case 'w':
				++availible[3];
				break;
			case 'e':
				++availible[4];
				break;
			case 'r':
				++availible[5];
				break;
		}
		
		if(!(mapFile >> nextObjName))
		{
			errors::mapLoadError(path, "Error in object-list");
			break;
		}
	}
	
	string lastLvlMsg; 
	if(lastLevel)
	{
		int fighters = totals.saved[0];
		if(totals.saved[2] < fighters)
		{
			fighters = totals.saved[2];
			infos->medkitsAvailible = 0;
		}
		else
		{
			infos->medkitsAvailible = totals.saved[2] - fighters;
		}
		totals.endFight = fighters;
		int goodFighters = fighters;
		if(totals.saved[3] < goodFighters)
		{
			goodFighters = totals.saved[3];
			infos->ammoAvailible = 0;
		}
		else
		{
			infos->ammoAvailible = totals.saved[3] - goodFighters;
		}
		lastLvlMsg = "You get "+i2str(fighters)+" allies, "+i2str(goodFighters)+" with good weapons.";
		
		const string allyIds[3] = { "ajunge", "aaltermann", "agrau" };
		cobject *allyTemps[3];
		for(int i = 0; i < 3; ++i)
		{
			allyTemps[i] = findTemplate(allyIds[i]);
		}
		for(int i = 0; i < fighters; ++i)
		{
			cobject *newAlly = new cobject();
			*newAlly = *(allyTemps[rand() % 3]);
			newAlly->setPos(rand() % 240 + 60, rand() % 552 + 24);
			if(i < goodFighters)
			{
				newAlly->setWeaponsAvailible(5);
			}
			objects.push_back(newAlly);
		}
	}
	
	mapFile >> introMovie;
	menu::showmovie(introMovie);
	
	menu::showBriefing(mapFile, targets, availible, timeLimit);
	if(lastLevel)
	{
		menu::showAchievement(4, lastLvlMsg);
	}
	
	mapFile.close();
}

void cmap::move(int dx, int dy)
{
	player->setPosX(player->getPosX() + dx);
	player->setPosY(player->getPosY() + dy);
}

objList &cmap::getObjects()
{
	return objects;
}

int cmap::tick()
{
	++infos->timeUsed;
	if(timeLeft > 0)
	{
		--timeLeft;
	}
	else
	{
		if(lastLevel)
		{
			return 0;
		}
		--clockColorCounter;
		if(clockColorCounter <= 0)
		{
			clockColorCounter = 60;
		}
	}
	if(player->getLife() <= 0)
	{
		return 2;
	}
	
	//Check for winning-condition-cheat
	if(key[KEY_LSHIFT] && keyStates[KEY_ESC])
	{
		return 0;
	}
	
	if(keyStates[KEY_ESC])
	{
		if(menu::showIngameMenu() != 0)
		{
			return 1;
		}
	}
	if(keyStates[global.keyMedipack] && player->getLife() < 50 && infos->medkitsAvailible > infos->medkitsUsed)
	{
		++infos->medkitsUsed;
		player->setLife(100);
	}
	
	
	if(keyStates[KEY_F1])
	{
		ifstream briefFile("media/movies/help.txt");
		menu::showBriefing(briefFile, 0, 0, 0);
		briefFile.close();
	}
	
	for(int i = 0; i < 5; ++i)
	{
		if(weapons[i].coolDown > 0)
		{
			--weapons[i].coolDown;
		}
	}
	
	if(timeLimit > 0 && !timeLeft)
	{
		if((clockColorCounter % 14) == 0)
		{
			int firePosX = (rand() % 180) - 90;
			int firePosY = (rand() % 180) - 70;
			if(firePosX < 0 && firePosX > -15)
			{
				firePosX -= 15;
			}
			else if(firePosX >= 0 && firePosX < 15)
			{
				firePosX += 15;
			}
			if(firePosY < 0 && firePosY > -10)
			{
				firePosY -= 10;
			}
			else if(firePosY >= 0 && firePosY < 15)
			{
				firePosY += 15;
			}
			firePosX += player->getPosX();
			firePosY += player->getPosY();
			
			if(firePosX >= 0 && firePosY >= 0)
			{
				if(!fireTemplate)
				{
					fireTemplate = findTemplate("rfire");
				}
				cobject *newFire = new cobject();
				*newFire = *fireTemplate;
				newFire->setPos(firePosX, firePosY);
				objects.push_back(newFire);
				++availible[4];
			}
		}
	}
	
	int dropzoneDistance = squareVal(player->getPosX() - entranceX) + squareVal(player->getPosY() - entranceY);
	if(leftDropzone)
	{
		if(dropzoneDistance < 24*24)
		{
			leftDropzone = false;
			bool needItems = false;
			for(int i = 0; i < 4; ++i)
			{
				if(targets[i] > infos->saved[i])
				{
					needItems = true;
					break;
				}
			}
			blit(global.doublebuffer, global.mainmenu, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
			masked_blit(menuBgTitle, global.mainmenu, 0, 0, (SCREEN_W - menuBgTitle->w) / 2, 10, menuBgTitle->w, menuBgTitle->h);
			
			const int BUTTONS = 2;
			string labels[BUTTONS] = { "Go back in!", "Leave this place..." };
			bool states[BUTTONS] = { true, !needItems };
			
			//print info on size 180/220
			rect(global.mainmenu, 180, 220, 620, 380, 0);
			rectfill(global.mainmenu, 181, 221, 619, 379, colors::white);
			int linePos = 224;
			int lineHeight = text_height(global.size12) + 3;
			if(needItems)
			{
				textout_ex(global.mainmenu, global.size12bold, "Wait! You're still missing some items!", 185, linePos, 0, -1);
				linePos += lineHeight;
				textout_ex(global.mainmenu, global.size12, "You still need to...", 185, linePos, 0, -1);
				linePos += lineHeight;
				for(int i = 0; i < 4; ++i)
				{
					const char *msgs[4] = { "    ...rescue %i people.", "    ...collect %i food rations.", "    ...collect %i medipacks.", "    ...collect %i weapons." };
					if(infos->saved[i] >= targets[i])
					{
						continue;
					}
					textprintf_ex(global.mainmenu, global.size12, 185, linePos, 0, -1, msgs[i], targets[i] - infos->saved[i]);
					linePos += lineHeight;
				}
			}
			else
			{
				textout_ex(global.mainmenu, global.size12bold, "Congratulations! You got the needed items.", 185, linePos, 0, -1);
				linePos += lineHeight;
				textout_ex(global.mainmenu, global.size12, "If you want, you can go back in and collect more.", 185, linePos, 0, -1);
				linePos += lineHeight;
				textout_ex(global.mainmenu, global.size12, "Or, you can just leave and go to the next market.", 185, linePos, 0, -1);
				linePos += lineHeight;
				if(timeLimit)
				{
					textout_ex(global.mainmenu, global.size12, "But be careful: You don't have much time!", 185, linePos, 0, -1);
					linePos += lineHeight;
				}
			}
			textout_ex(global.mainmenu, global.size12, "(All inventory items have been given to Leonardo.)", 185, linePos, 0, -1);
			
			
			inventoryItems = 0;
			if(menu::stdMenu(0, BUTTONS, labels, states, global.size20, 75, 0, 0, global.mainmenu, 0))
			{
				return 0;
			}
		}
	}
	else if(dropzoneDistance > 24*24)
	{
		leftDropzone = true;
	}
	
	for(objListIter ter = objects.begin(); ter != objects.end(); ++ter)
	{
		if((*ter)->getLife() <= 0 && (*ter) != player)
		{
			if((*ter)->getAiType() == 2)
			{
				++infos->killed[1];
			}
			else if((*ter)->getAiType() > 2 && (*ter)->getAiType() < 7)
			{
				++infos->killed[0];
			}
			delete (*ter);
			ter = objects.erase(ter);
			--ter;
			continue;
		}
		(*ter)->tick(*this);
	}
	
	return -1;
}

cobject *cmap::getPlayer()
{
	return player;
}

void cmap::spawnCreature(int posX, int posY, int type)
{
	cobject *newEnemy = new cobject();
	const string enemyNames[4] = { "epig", "egolem", "edragon", "ebasher" };
	*newEnemy = *findTemplate(enemyNames[type]);
	
	newEnemy->setPos(posX, posY);
	++infos->availibleKills[0];
	
	objects.push_back(newEnemy);
}

void cmap::draw()
{
	//draw background
	const int GUISIZE = 100;
	int playerX = player->getPosX(), playerY = player->getPosY();
	int camX = playerX - SCREEN_W / 2,
		camY = playerY - SCREEN_H / 2;
	if(camX < 0)
	{
		camX = 0;
	}
	if(camY < 0)
	{
		camY = 0;
	}
	if(camX > sizeX - SCREEN_W)
	{
		camX = sizeX - SCREEN_W;
	}
	if(camY > sizeY - SCREEN_H)
	{
		camY = sizeY - SCREEN_H;
	}
	int floorTilesX = (SCREEN_W) / floorSizeX + 2;
	int floorTilesY = SCREEN_H / floorSizeY + 2;
	int minusX = camX % floorSizeX;
	int minusY = camY % floorSizeY;
	for(int x = 0; x < floorTilesX; ++x)
	{
		for(int y = 0; y < floorTilesY; ++y)
		{
			blit(tilesetGfx, global.doublebuffer, floorX, floorY, x * floorSizeX - minusX, y * floorSizeY - minusY, floorSizeX, floorSizeY);
		}
	}
	masked_blit(entryGfx, global.doublebuffer, 0, 0, entranceX-16 - camX, entranceY - 8 - camY, 16, 16);
	masked_blit(entryGfx, global.doublebuffer, 0, 0, entranceX - camX, entranceY - 8 - camY, 16, 16);
	
	//Sort object-list by y value ("z-buffer")
	objects.sort(sortCObjects);
	
	//draw units
	for(objListIter ter = objects.begin(); ter != objects.end(); ++ter)
	{
		(*ter)->draw(camX, camY);
	}
	
	//Draw GUI
	if(timeLimit > 0)
	{
		circlefill(global.doublebuffer, 0, 0, 100, 0);
		if(timeLeft > 0)
		{
			int timeTmp = timeLeft;
			int minutes = timeTmp / 60 / 60;
			timeTmp -= minutes * 60 * 60;
			int seconds = timeTmp / 60;
			timeTmp -= seconds * 60;
			int hundreds = timeTmp * 100 / 60;
			line(global.doublebuffer, 0, 0, int(100.0 * std::sin(float(hundreds) / 100.0 * dPI)), int(100.0 * std::cos(float(hundreds) / 100.0 * dPI)), colors::green);
			textprintf_ex(global.doublebuffer, global.size12bold, 5, 20, colors::green, -1, "%i:%2i:%2i", minutes, seconds, hundreds);
		}
		else
		{
			textout_ex(global.doublebuffer, global.size12bold, "0:00:00", 5, 20, clockColorCounter > 30 ? colors::red : colors::green, -1);
		}
		//textprintf_ex(global.doublebuffer, global.size8, 3, 3, colors::white, 0, "%i/%i %i/%i", (*player)->getPosX(), (*player)->getPosY(), camX, camY);
	}
	//Life-Bar
	int lr = 0;
	int lg = 255;
	if(player->getLife() > 50)
	{
		lr = 255 - int(255.0f * float(player->getLife() - 50) / 50.0f);
	}
	else
	{
		lr = 255;
		lg = int(255.0f * float(player->getLife()) / 50.0f);
	}
	rect(global.doublebuffer, 692, 5, 794, 22, 0);
	rectfill(global.doublebuffer, 793 - player->getLife(), 6, 793, 21, makecol(lr, lg, 0));
	rectfill(global.doublebuffer, 692, 23, 794, 45, 0);
	draw_sprite(global.doublebuffer, icons[0], 700, 24);
	textprintf_ex(global.doublebuffer, global.size12, 735, 24, colors::white, -1, " x %i", infos->medkitsAvailible - infos->medkitsUsed);
	//Weapon
	int wep = player->getWeapon();
	rectfill(global.doublebuffer, 0, 559, 100, 599, colors::white);
	vline(global.doublebuffer, 25, 559, 599, 0);
	lr = 0;
	lg = 255;
	float ammoOutOf1 = float(globalAmmo) / float(AMMOVALUE);
	if(ammoOutOf1 > 0.5f)
	{
		lr = 255 - int(255.0f * (ammoOutOf1 - 0.5f) / 0.5f);
	}
	else
	{
		lr = 255;
		lg = int(255.0f * ammoOutOf1 / 0.5f);
	}
	rectfill(global.doublebuffer, 0, 599 - int(ammoOutOf1 * 40.0f), 24, 599, makecol(lr, lg, 0));
	if(weapons[wep].gfx)
	{
		masked_blit(weapons[wep].gfx, global.doublebuffer, 24, 0, 2, 565, 24, 32);
	}
	//textprintf_ex(global.doublebuffer, global.size12, 0, 570, colors::red, key[KEY_SPACE] ? colors::green : -1, "%i", globalAmmo);
	draw_sprite(global.doublebuffer, icons[1], 27, 560);
	textprintf_ex(global.doublebuffer, global.size12, 65, 570, 0, -1, "x %i", infos->ammoAvailible - infos->ammoUsed);
	
	//Inventory
	rectfill(global.doublebuffer, 749, 579, 799, 599, colors::white);
	textout_centre_ex(global.doublebuffer, global.size8, "Inventory", 774, 579, 0, -1);
	textprintf_centre_ex(global.doublebuffer, global.size8, 774, 589, 0, -1, "%i / %i", inventoryItems, MAXINVENTORY);
}

void cmap::shootBullet(int startX, int startY, int dir, cobject *shooter)
{
	if(shooter == player)
	{
		++infos->bulletsUsed;
	}
	int bulletX = startX, bulletY = startY;
	cobject *collision = 0;
	int shooterWeapon = shooter->getWeapon();
	int maxDistance = shooterWeapon == 1 ? 70 : 500;
	while(!collision)
	{
		if(bulletX < 0 || bulletX > sizeX || bulletY < 0 || bulletY > sizeY)
		{
			break;
		}
		if(std::abs(bulletX - startX) > maxDistance || std::abs(bulletY - startY) > maxDistance)
		{
			break;
		}
		for(objListIter ter = objects.begin(); ter != objects.end(); ++ter)
		{
			cobject *curObj = *ter;
			if(curObj == shooter || curObj->getWalkable() || curObj->getAiType() == 1 || curObj->getAiType() > 6 || (shooterWeapon != 1 && curObj->getAiType() == 2))
			{
				continue;
			}
			if(collides(bulletX - 2, bulletY - 2, bulletX + 2, bulletY + 2,
				curObj->getPosX() + curObj->getCollision()[3], curObj->getPosY() + curObj->getCollision()[0], curObj->getPosX() + curObj->getCollision()[1], curObj->getPosY() + curObj->getCollision()[2]))
			{
				collision = curObj;
			}
			if(collision)
			{
				break;
			}
		}
		bulletX += 12 * moveDelta[dir][0];
		bulletY += 12 * moveDelta[dir][1];
	}
	if(collision)
	{
		if(shooterWeapon != 1 && collision->getAiType() > 2 && collision->getAiType() < 7)
		{
			collision->substractLife(weapons[shooterWeapon].dmg);
			if(shooter == player)
			{
				++infos->bulletsHit;
			}
		}
		else if(shooterWeapon == 1 && collision->getAiType() == 2)
		{
			collision->substractLife(weapons[shooterWeapon].fireDmg);
			++infos->bulletsHit;
		}
	}
}

void cmap::pickupObject(cobject *object)
{
	char objId = object->getId()[0];
	int invSpaces = (objId == 'h') ? 3 : 1;
	if(invSpaces + inventoryItems <= INVENTORYSIZE)
	{
		inventoryItems += invSpaces;
		if(objId == 'h')
		{
			++infos->saved[0];
		}
		else if(objId == 'f')
		{
			++infos->saved[1];
		}
		else if(objId == 'm')
		{
			++infos->saved[2];
		}
		else if(objId == 'w')
		{
			++infos->saved[3];
		}
		string followupId = object->getFollowupId();
		if(followupId != "0")
		{
			cobject *followuptmp = findTemplate(followupId);
			*object = *followuptmp;
		}
		else
		{
			object->setLife(0);
		}
	}
}

bool cmap::useAmmo()
{
	if(infos->ammoAvailible > infos->ammoUsed)
	{
		globalAmmo = AMMOVALUE;
		++infos->ammoUsed;
		return true;
	}
	return false;
}

/*void cmap::drawOverview(BITMAP *targetBmp, int targetPosX, int targetPosY, int targetSizeX, int targetSizeY)
{
	rectfill(targetBmp, targetPosX, targetPosY, targetPosX + targetSizeX, targetPosY + targetSizeY, makecol(200, 200, 250));
	
	for(objListIter ter = objects.begin(); ter != objects.end(); ++ter)
	{
		cobject *curObj = *ter;
		if(!curObj->getWalkable())
		{
			//
		}
	}
}*/

cobject *cmap::checkCollision(cobject *who, int newPosX, int newPosY)
{
	cobject *collision = 0;
	
	if(newPosX + who->getCollision()[3] < 0)
	{
		return who;
	}
	else if(newPosY + who->getCollision()[0] < 0)
	{
		return who;
	}
	else if(newPosX + who->getCollision()[1] >= sizeX)
	{
		return who;
	}
	else if(newPosY + who->getCollision()[2] >= sizeY)
	{
		return who;
	}
	
	for(objListIter ter = objects.begin(); ter != objects.end(); ++ter)
	{
		if(*ter == who || (*ter)->getWalkable())
		{
			continue;
		}
		cobject *curObj = *ter;
		if(collides(newPosX + who->getCollision()[3], newPosY + who->getCollision()[0], newPosX + who->getCollision()[1], newPosY + who->getCollision()[2],
			curObj->getPosX() + curObj->getCollision()[3], curObj->getPosY() + curObj->getCollision()[0], curObj->getPosX() + curObj->getCollision()[1], curObj->getPosY() + curObj->getCollision()[2]))
		{
			collision = curObj;
		}
		if(collision)
		{
			break;
		}
	}
	return collision;
}

bool cmap::isFirefighter()
{
	return firefighterUnlocked;
}

weapon weapons[5];
BITMAP *weaponFire[2] = {0,0};
int globalAmmo = 0;

weapon::weapon()
{
	gfx = 0;
	ammo = 0;
	coolDown = 0;
}

weapon::~weapon()
{
}

void weapon::unload()
{
	if(gfx)
	{
		destroy_bitmap(gfx);
		gfx = 0;
	}
}

void weapon::load(int nid)
{
	id = nid + 1;
	ifstream readFile(("media/weapons/" + i2str(id) + ".txt").c_str());
	
	if(!readFile)
	{
		return;
	}
	
	string path;
	readFile >> name >> path >> dmg >> fireDmg >> fireRate;
	gfx = loadBitmap(path);
}
