/*
  Name: object.cpp
  Author: Master Lenman
  Project: SpeedHack 2009
  Date: 10.07.2009
  Description: Base class for all objects
*/

#include <allegro.h>
#include <string>
#include <fstream>
#include <cmath>

#include "cmap.h"
#include "cobject.h"
#include "global.h"
#include "main.h"
#include "errors.h"
#include "keyStatesClass.h"

using namespace std;

const int moveDelta[4][2] = { { 0, -1 }, { 1, 0 }, { 0, 1 }, { -1, 0 } };

cobject::cobject()
{
	gfx = 0;
	posX = posY = 0;
	destroyGfx = false;
	gfxPosX = gfxPosY = 0;
	sizeX = sizeY = 0;
	for(int i = 0; i < 4; ++i)
	{
		collision[i] = 0;
	}
	walkable = false;
	collectable = false;
	characterAnim = false;
	moving = false;
	dir = 2;
	moveDir = 2;
	curAnimation = 1;
	animationCounter = 0;
	spawnType = spawnFrequency = 0;
	weapon = 0;
	weaponsAvailible = 2;
	speed = 2;
	aiObstacle = aiRoaming = 0;
	life = maxLife = 0;
	weaponCooldown = 0;
	drawBullet = 0;
}

cobject::~cobject()
{
	if(gfx && destroyGfx)
	{
		destroy_bitmap(gfx);
	}
}

cobject &cobject::operator=(const cobject &temp)
{
	bool followup = (followupId == temp.getId());
	//allegro_message(("copying" + temp.getId()).c_str());
	id = temp.getId();
	followupId = temp.getFollowupId();
	if(!followup)
	{
		posX = temp.getPosX();
		posY = temp.getPosY();
	}
	for(int i = 0; i < 4; ++i)
	{
		collision[i] = temp.getCollision()[i];
	}
	gfx = temp.getGfx();
	gfxPosX = temp.getGfxPosX();
	gfxPosY = temp.getGfxPosY();
	sizeX = temp.getSizeX();
	sizeY = temp.getSizeY();
	walkable = temp.getWalkable();
	collectable = temp.getCollectable();
	characterAnim = temp.getCharacterAnim();
	aiType = temp.getAiType();
	spawnType = temp.getSpawnType();
	spawnFrequency = temp.getSpawnFrequency();
	life = maxLife = temp.getLife();
	destroyGfx = false;
	speed = temp.getSpeed();
	
	return *this;
}

bool sortCObjects(cobject *a, cobject *b)
{
	return *a < *b;
}

bool cobject::operator<(const cobject &temp)
{
	int top = temp.getPosY() - temp.getSizeY() / 2;
	return (posY - sizeY / 2) < top;
}

//retVals:
	//0 = do nothing
	//1 = delete me! (for dead monsters/ai fighters)
int cobject::tick(cmap &theMap)
{
	if(!aiType)
	{
		return 0;
	}
	int retVal = 0;
	if(drawBullet < 0)
	{
		++drawBullet;
	}
	else if(drawBullet > 0)
	{
		--drawBullet;
	}
	if(aiObstacle > 0)
	{
		--aiObstacle;
	}
	if(weaponCooldown)
	{
		--weaponCooldown;
	}
	if(aiType == 8)
	{
		if(!weaponCooldown)
		{
			//Spawn enemy!
			weaponCooldown = spawnFrequency;
			--life;
			
			theMap.spawnCreature(posX, posY, spawnType);
		}
	}
	if(characterAnim)
	{
		if(aiType == 2)
		{
			++animationCounter;
			if((animationCounter % 3) == 0)
			{
				++curAnimation;
				if(curAnimation == 3)
				{
					curAnimation = 0;
					++dir;
					dir %= 4;
				}
			}
		}
		int dX = 0, dY = 0;
		if(aiType == 1)
		{
			if(key[global.keyLeft])
			{
				moving = true;
				moveDir = 3;
			}
			else if(key[global.keyRight])
			{
				moving = true;
				moveDir = 1;
			}
			else if(key[global.keyUp])
			{
				moving = true;
				moveDir = 0;
			}
			else if(key[global.keyDown])
			{
				moving = true;
				moveDir = 2;
			}
			else
			{
				moving = false;
				animationCounter = 0;
			}
			if(key[global.keyShoot] && !weaponCooldown)
			{
				if(globalAmmo >= weapons[weapon].dmg || weapon < 2)
				{
					//pistol doesn't use ammo
					if(weapon > 1)
					{
						globalAmmo -= weapons[weapon].dmg;
						if(!globalAmmo)
						{
							if(!theMap.useAmmo())
							{
							//	keyStates.setState(global.keyChangeWeapon, true);
							}
						}
					}
					theMap.shootBullet(posX, posY, dir, this);
					drawBullet = 2;
					if(weapons[weapon].fireDmg > 0)
					{
						drawBullet = -9;
					}
					weaponCooldown = weapons[weapon].fireRate;
				}
				else
				{
					if(!theMap.useAmmo())
					{
						keyStates.setState(global.keyChangeWeapon, true);
					}
				}
			}
			if(keyStates[global.keyChangeWeapon])
			{
				do
				{
					++weapon;
					weapon %= weaponsAvailible;
				} while(!(weapon < 2 || globalAmmo >= weapons[weapon].dmg || theMap.useAmmo()));
			}
			if(!key[global.keyStrafe])
			{
				dir = moveDir;
			}
		}
		//allies
		else if(aiType == 7)
		{
			if(!weaponCooldown)
			{
				theMap.shootBullet(posX, posY, dir, this);
				drawBullet = 2;
				weaponCooldown = weapons[weapon].fireRate;
			}
			moving = true;
			dir = 1;
			if(aiRoaming == -1)
			{
				aiRoaming = 120 + rand() % 25;
				moveDir = (rand() % 2) * 2;
			}
			else if(aiRoaming == 0)
			{
				aiRoaming = -30 - rand() % 15;
				moveDir = (rand() % 2) * 2 + 1;
			}
			if(aiRoaming < -1)
			{
				++aiRoaming;
				if((posX <= 60 && dir == 3) || (posX >= 300 && dir == 1))
				{
					aiRoaming = -1;
				}
			}
			else
			{
				--aiRoaming;
				if((posY <= 24 && dir == 0) || (posY >= 576 && dir == 2))
				{
					aiRoaming = 0;
				}
			}
		}
		//enemies!
		else if(aiType > 2 && aiType < 7)
		{
			cobject *player = theMap.getPlayer();
			dX = player->getPosX() - posX;
			dY = player->getPosY() - posY;
			moving = true;
			if(int(std::abs(dX) > std::abs(dY)) ^ int((aiObstacle > 0)))
			{
				if(dX > 0)
				{
					dir = moveDir = 1;
				}
				else
				{
					dir = moveDir = 3;
				}
			}
			else
			{
				if(dY > 0)
				{
					dir = moveDir = 2;
				}
				else
				{
					dir = moveDir = 0;
				}
			}
		}
		if(moving)
		{
			++animationCounter;
			
			if((animationCounter % speed) == 0 || speed == 0 || speed == 1)
			{
				int limit = 1;
				if(speed == 0)
				{
					limit = 2;
				}
				
				for(int i = 0; i < limit; ++i)
				{
					int newPosX = posX + moveDelta[moveDir][0];
					int newPosY = posY + moveDelta[moveDir][1];
					//check for collisions, if no collisions:
					cobject *collision = theMap.checkCollision(this, newPosX, newPosY);
					if(collision && collision != this)
					{
						if(aiType == 1 && collision->getAiType() == 2)
						{
							life -= 3;
							if(theMap.isFirefighter())
							{
								life += 2;
							}
						}
						else if((aiType == 1 || aiType == 7) && (collision->getAiType() == 7 || collision->getAiType() == 1))
						{
							collision = false;
						}
						else if(aiType == 1 && collision->getCollectable())
						{
							theMap.pickupObject(collision);
						}
						else if(aiType > 2 && aiType < 7)
						{
							if(collision->getAiType() > 1 && collision->getAiType() < 7)
							{
								collision = false;
							}
							else if(collision->getAiType() == 1 || collision->getAiType() == 7)
							{
								collision->substractLife(aiType - 2);
							}
							else if(collision->getAiType() == 0)
							{
								//if((dir & 1) && 
								aiObstacle = 20;
							}
							
						}
					}
					if(!collision)
					{
						posX = newPosX;
						posY = newPosY;
					}
					else
					{
						break;
					}
				}
			}
			
			if(curAnimation == 1)
			{
				curAnimation = 0;
			}
			if((animationCounter % 12) == 0)
			{
				if(curAnimation == 0)
				{
					curAnimation = 2;
				}
				else if(curAnimation == 2)
				{
					curAnimation = 0;
				}
			}
		}
		else if(aiType != 2)
		{
			animationCounter = 0;
			curAnimation = 1;
		}
	}
	return retVal;
}

void cobject::drawWeapon(int destX, int destY)
{
	masked_blit(weapons[weapon].gfx, global.doublebuffer, dir * 24, 0, destX, destY, 24, 32);
	destX += 12;
	destY += 16;
	if(drawBullet > 0)
	{
		destX += moveDelta[dir][0] * 12;
		destY += moveDelta[dir][1] * 16;
		rotate_sprite(global.doublebuffer, weaponFire[0], destX, destY, itofix(64 * dir));
	}
	else if(drawBullet < 0)
	{
		destX += moveDelta[dir][0] * 20;
		destY += moveDelta[dir][1] * 24;
		rotate_sprite(global.doublebuffer, weaponFire[1], destX, destY, itofix(64 * dir));
	}
}

void cobject::draw(int camX, int camY)
{
	if(gfx)
	{
		int destX = posX - camX - sizeX / 2;
		int destY = posY - camY - sizeY / 2;
		if(!characterAnim)
		{
			masked_blit(gfx, global.doublebuffer, gfxPosX, gfxPosY, destX, destY, sizeX, sizeY);
		}
		else
		{
			if((aiType == 1 || aiType == 7) && (dir == 0 || dir == 3))
			{
				drawWeapon(destX, destY);
			}
			masked_blit(gfx, global.doublebuffer, gfxPosX + sizeX * curAnimation, gfxPosY + dir * sizeY, destX, destY, sizeX, sizeY);
			if((aiType == 1 || aiType == 7) && (dir == 1 || dir == 2))
			{
				drawWeapon(destX, destY);
			}
		}
		if(drawBullet < 0)
		{
		}
	}
}

void cobject::loadTemplate(string oid, string path)
{
	ifstream readfile(path.c_str());
	if(!readfile)
	{
		errors::mapLoadError(path, "Cannot open object file!");
	}
	
	string fileId;
	readfile >> fileId;
	if(oid != fileId)
	{
		errors::mapLoadError(path, "Object load error - wrong id!");
	}
	//allegro_message(("loading"+oid).c_str());
	id = oid;
	
	string tilesetPath;
	readfile >> tilesetPath;
	gfx = loadBitmap(tilesetPath);
	
	int tilesetCoords[4];
	for(int i = 0; i < 4; ++i)
	{
		readfile >> tilesetCoords[i];
	}
	gfxPosX = tilesetCoords[0];
	gfxPosY = tilesetCoords[1];
	sizeX = std::abs(tilesetCoords[2] - tilesetCoords[0]) + 1;
	sizeY = std::abs(tilesetCoords[3] - tilesetCoords[1]) + 1;
	
	for(int i = 0; i < 4; ++i)
	{
		readfile >> collision[i];
	}
	
	readfile >> walkable >> collectable;
	readfile >> characterAnim >> aiType >> life >> speed;
	readfile >> spawnType >> spawnFrequency;
	readfile >> followupId;
	
	readfile.close();
}

string cobject::getId() const
{
	return id;
}
void cobject::setWeaponsAvailible(int nAvai)
{
	weaponsAvailible = nAvai;
	if(weaponsAvailible > 2)
	{
		weapon = weaponsAvailible - 1;
	}
}

string cobject::getFollowupId() const
{
	return followupId;
}

int cobject::getSpeed() const
{
	return speed;
}

int cobject::getLife() const
{
	return life;
}

void cobject::substractLife(int dLife)
{
	life -= dLife;
}

void cobject::setLife(int nLife)
{
	life = nLife;
}

int cobject::getPosX() const
{
	return posX;
}

int cobject::getPosY() const
{
	return posY;
}

int cobject::getWeapon() const
{
	return weapon;
}

void cobject::setPosX(int newX)
{
	posX = newX;
}

void cobject::setPosY(int newY)
{
	posY = newY;
}

void cobject::setPos(int newX, int newY)
{
	posX = newX;
	posY = newY;
}

const int *cobject::getCollision() const
{
	return collision;
}

BITMAP *cobject::getGfx() const
{
	return gfx;
}

int cobject::getGfxPosX() const
{
	return gfxPosX;
}

int cobject::getGfxPosY() const
{
	return gfxPosY;
}

int cobject::getSizeX() const
{
	return sizeX;
}

int cobject::getSizeY() const
{
	return sizeY;
}

bool cobject::getWalkable() const
{
	return walkable;
}

bool cobject::getCollectable() const
{
	return collectable;
}

bool cobject::getCharacterAnim() const
{
	return characterAnim;
}

int cobject::getAiType() const
{
	return aiType;
}

int cobject::getSpawnType() const
{
	return spawnType;
}

int cobject::getSpawnFrequency() const
{
	return spawnFrequency;
}
