/*
  Name: Errors.cpp
  Author: Master Lenman
  Project: SpeedHack 2009
  Date: 09.07.2009
  Description: This file contains the functions of the errors namespace,
	  that will show error messages for all kind of errors
*/

#include <string>
#include <allegro.h>

#include "Errors.h"

using namespace std;

void errors::fatalError(string message, bool kill)
{
	if(kill)
	{
		set_gfx_mode(GFX_TEXT, 0,0,0,0);
    	allegro_message(message.c_str());
	    exit(0);
	}
	else
	{
		string line2, line3;
		int position = message.find('\n');
		if(position != string::npos)
		{
			line2 = message.c_str() + position + 1;
			message.erase(position);
			position = line2.find('\n');
			if(position != string::npos)
			{
				line3 = line2.c_str() + position + 1;
				line2.erase(position);
			}
		}
		alert(message.c_str(), line2.c_str(), line3.c_str(), "OK", 0, 0, 0);
	}
}

void errors::timerError(bool kill)
{
	fatalError("Error001: Timer could not be initialised", kill);
}

void errors::keyboardError(bool kill)
{
	fatalError("Error002: No keyboarddriver found", kill);
}

void errors::soundDriverError(bool kill)
{
	fatalError("Error003: Sound driver could not be initialised", kill);
}

void errors::graphicModeError(bool kill)
{
	fatalError("Error004: Graphic mode couldn't be set", kill);
}

void errors::createError(bool kill)
{
	fatalError("Error005: Creating bitmap failed", kill);
}

void errors::loadBitError(string name, bool kill)
{
	fatalError(string("Error006: Loading bitmap \"") + name + string("\" failed"), kill);
}

void errors::datafileError(string name, bool kill)
{
	fatalError("Error007: Loading datafile \"" + name + "\" failed", kill);
}

void errors::sampleLoadError(string sampleName, bool kill)
{
	fatalError("Error008: Loading sample \"" + sampleName + "\" failed", kill);
}

void errors::divBy0Error(string module, bool kill)
{
	fatalError("Error009: A division by 0 error was prevented in the following module:\n" + module, kill);
}

void errors::mapLoadError(string mapName, string reason, bool kill)
{
	fatalError("Error010: Could not load map \"" + mapName + "\" for the following reason:\n" + reason, kill);
}

void errors::mouseError(bool kill)
{
	fatalError("Error011: Failed to initialise mouse", kill);
}
	
//Error in network code
void errors::networkError(string errorMsg, bool kill)
{
	fatalError("Error012: Network error. Error Message:\n" + errorMsg, kill);
}

void errors::serverConnectError(std::string errorMsg, bool kill)
{
	fatalError("Error013: The server refused the connection. Message by server:\n" + errorMsg, kill);
}
