/*
  Name: game.cpp
  Author: Master Lenman
  Project: SpeedHack 2009
  Date: 10.07.2009
  Description: 
*/

#include <allegro.h>

#include "game.h"
#include "main.h"
#include "keyStatesClass.h"
#include "global.h"
#include "cmap.h"
#include "iniReader.h"
#include "menues.h"

using namespace std;

int limitCarrySize(int availible)
{
	if(availible > 3)
	{
		if(availible > 9)
		{
			availible = availible / 3 + 2; 
		}
		else if(availible > 5)
		{
			availible = 4;
		}
		else
		{
			availible = 3;
		}
	}
	return availible;
}

const int UNLOCKSTAGES[3] = { 3, 8, 14 };
const int LASTLEVEL = 8;
void startGame(bool loadGame)
{
	int mapId = 1;
	int weaponsUnlocked = 2;
	bool firefighterUnlocked = false;
	string nextMap;
	for(int i = 0; i < 5; ++i)
	{
		weapons[i].ammo = 0;
	}
	
	globalAmmo = AMMOVALUE;
	gameInfos totals;
	//give the player 2 medkits as a start
	totals.saved[2] = 2;
	totals.availibleSaves[2] = 2;
	
	if(loadGame)
	{
		totals.load();
		weaponsUnlocked = totals.weaponsUnlocked;
		firefighterUnlocked = totals.firefighterUnlocked;
		mapId = totals.level;
	}
	//TODO: remove the next line ;)
//	mapId = 8;
//	weaponsUnlocked = 5;
	
	//gameOver states:
		//0 = load next map!
		//1 = player quit himself
		//2 = player died -> gameover!
		//3 = player finished the game! Wooohoo!
	int mapResult;
	while(mapId <= LASTLEVEL)
	{
		bool lastLevel = (mapId == LASTLEVEL);
		nextMap = "media/maps/map" + i2str(mapId) + ".txt";
		//Save game
		totals.level = mapId;
		totals.weaponsUnlocked = weaponsUnlocked;
		totals.firefighterUnlocked = firefighterUnlocked;
		totals.save();
		
		//start next level
		gameInfos levelInfos;
		levelInfos.medkitsAvailible = limitCarrySize(totals.saved[2]);
		levelInfos.ammoAvailible = limitCarrySize(totals.saved[3]);
		mapResult = runMap(nextMap, &levelInfos, weaponsUnlocked, firefighterUnlocked, lastLevel, totals);
		
		if(lastLevel && !mapResult)
		{
			menu::showLevelResults(totals, true);
			menu::showEndResult(totals);
			break;
		}
		
		totals.add(levelInfos);
		
		menu::showLevelResults(levelInfos, false);
		menu::showLevelResults(totals, true, mapResult == 2);
		
		//If total ammo-count is bigger than X, the next weapon is unlocked!
		for(int i = 0; i < 3; ++i)
		{
			if(weaponsUnlocked > i + 2)
				continue;
			if(totals.saved[3] > UNLOCKSTAGES[i])
			{
				weaponsUnlocked = 3 + i;
				menu::showAchievement(i);
			}
		}
		if(!firefighterUnlocked)
		{
			if(float(levelInfos.killed[1]) / float(levelInfos.availibleKills[1]) > 0.85f)
			{
				firefighterUnlocked = true;
				menu::showAchievement(3);
			}
		}
		
		if(mapResult)
		{
			break;
		}
		++mapId;
	}
}

int runMap(string mapId, gameInfos *levelInfos, int weaponsUnlocked, bool firefighterUnlocked, bool lastLevel, gameInfos &totals)
{
	cmap map;
	map.load(mapId, levelInfos, weaponsUnlocked, firefighterUnlocked, lastLevel, totals);
	timerCounter = 0;
	bool run = true;
	int retVal = 0;
	while(run)
	{
		while(timerCounter > 0)
		{
			retVal = map.tick();
			if(retVal > -1)
			{
				run = false;
				break;
			}
			if(keyStates[KEY_ESC])
			{
				run = false;
				break;
			}
			
			--timerCounter;
		}
		
		clear_to_color(global.doublebuffer, 0);
		map.draw();
		show();
	}
	
	return retVal;
}

gameInfos::gameInfos()
{
	for(int i = 0; i < 4; ++i)
	{
		saved[i] = availibleSaves[i] = 0;
	}
	for(int i = 0; i < 2; ++i)
	{
		killed[i] = availibleKills[i] = 0;
	}
	endFightSurvivors = endFight = 0;
	bulletsUsed = bulletsHit = 0;
	medkitsUsed = medkitsAvailible = 0;
	ammoUsed = ammoAvailible = 0;
	
	level = 0;
	weaponsUnlocked = 2;
	firefighterUnlocked = false;
	timeUsed = 0;
}

gameInfos::~gameInfos()
{
}

void gameInfos::add(gameInfos &from)
{
	for(int i = 0; i < 4; ++i)
	{
		saved[i] += from.saved[i];
		availibleSaves[i] += from.availibleSaves[i];
	}
	for(int i = 0; i < 2; ++i)
	{
		killed[i] += from.killed[i];
		availibleKills[i] += from.availibleKills[i];
	}
	endFightSurvivors += from.endFightSurvivors;
	endFight += from.endFight;
	bulletsUsed += from.bulletsUsed;
	bulletsHit += from.bulletsHit;
	timeUsed += from.timeUsed / 60;
}

void gameInfos::save()
{
	iniWriter saveFile;
	
	saveFile.addEntry("level", level);
	saveFile.addEntry("globalAmmo", globalAmmo);
	saveFile.addEntry("weaponsUnlocked", weaponsUnlocked);
	saveFile.addEntry("firefighterUnlocked", firefighterUnlocked);
	saveFile.addEntry("timeUsed", timeUsed);
	
	saveFile.addCategory("statistics");
	for(int i = 0; i < 4; ++i)
	{
		saveFile.addEntry("saved" + i2str(i), saved[i]);
		saveFile.addEntry("availibleS" + i2str(i), availibleSaves[i]);
	}
	for(int i = 0; i < 2; ++i)
	{
		saveFile.addEntry("killed" + i2str(i), killed[i]);
		saveFile.addEntry("availibleK" + i2str(i), availibleKills[i]);
	}
	saveFile.addEntry("bulletsUsed", bulletsUsed);
	saveFile.addEntry("bulletsHit", bulletsHit);
	
	saveFile.save("media/savegame.ini");
}

void gameInfos::load()
{
	iniReader saveFile;
	saveFile.load("media/savegame.ini");
	
	level = saveFile.getIValue("level", 1);
	globalAmmo = saveFile.getIValue("globalAmmo", AMMOVALUE);
	weaponsUnlocked = saveFile.getIValue("weaponsUnlocked", 2);
	firefighterUnlocked = saveFile.getBValue("firefighterUnlocked", false);
	timeUsed = saveFile.getIValue("timeUsed", 0);
	
	saveFile.setCategory("statistics");
	for(int i = 0; i < 4; ++i)
	{
		saved[i] = saveFile.getIValue("saved" + i2str(i), 0);
		availibleSaves[i] = saveFile.getIValue("availibleS" + i2str(i), 0);
	}
	for(int i = 0; i < 2; ++i)
	{
		killed[i] = saveFile.getIValue("killed" + i2str(i), 0);
		availibleKills[i] = saveFile.getIValue("availibleK" + i2str(i), 0);
	}
	bulletsUsed = saveFile.getIValue("bulletsUsed", 0);
	bulletsHit = saveFile.getIValue("bulletsHit", 0);
}

bool collides(int left1, int top1, int right1, int bottom1, int left2, int top2, int right2, int bottom2)
{
	bool collides[4] = { false, false, false, false };

	//Left side colliding
	if(left1 <= right2 && left1 >= left2)
	{
		collides[0] = true;
	}
	//Right side colliding
	if(right1 >= left2 && right1 <= right2)
	{
		collides[1] = true;
	}
	//Top side colliding
	if(top1 <= bottom2 && top1 >= top2)
	{
		collides[2] = true;
	}
	//Bottom side colliding
	if(bottom1 >= top2 && bottom1 <= bottom2)
	{
		collides[3] = true;
	}

	if(!(collides[0] || collides[1] || collides[2] || collides[3]))
		return false;

	//Check total collision for all 4 corners
	if(collides[0] && collides[2])
		return true;
	if(collides[0] && collides[3])
		return true;
	if(collides[1] && collides[2])
		return true;
	if(collides[1] && collides[3])
		return true;

	if(bottom1 - top1 > bottom2 - top2)
	{
		if((collides[0] || collides[1]) && top1 < top2 && bottom1 > bottom2)
			return true;
	}
	if(right1 - left1 > right2 - left2)
	{
		if((collides[2] || collides[3]) && left1 < left2 && right1 > right2)
			return true;
	}

	return false;
}
