/*
  Name: main.cpp
  Author: Master Lenman
  Project: SpeedHack 2009
  Date: 09.07.2009
  Description: This file contains basic functions which will
	  do things like setting Allegro up, showing the title screen
	  or blit the doubleBuffer to the screen.
*/

#include <allegro.h>
#include <string>
#include <sstream>
#include <fstream>
#include <ctime>

#include "keyStatesClass.h"
#include "main.h"
#include "errors.h"
#include "global.h"
#include "menues.h"
//#include "mouse.h"
#include "game.h"

using namespace std;

//Version information
const string VERSIONSTRING = string("Shopping in Hell v0.1, built ") + string(__DATE__);

const int TIMER_FPS = 60;
const int SCREENW = 800, SCREENH = 600;
const double dPI = 3.14159265358979323846;
const float fPI = 3.14159265358979323846f;

void init();

//The main() function is called first when the program is executed
int main(int argc, char *argv[])
{
	init();
#ifndef __DEBUG__
	//menu::intro();
#endif
	
	int mode = 0;
	
	while((mode = menu::main(mode)) != 3)
	{
		if(mode == 0 || mode == 1)
		{
			//menu::showmovie("media/movies/scene1.txt");
			startGame(mode == 1);
		}
		else if(mode == 2)
		{
		}
	}
	global.unload();
	
	return 0;
} END_OF_MAIN()

//timerCounterUpdater() will update timerCounter 32 times per second.
//This timer is used to make the game run at the same speed on all computers.
volatile int timerCounter = 0;
void timerCounterUpdater()
{
	++timerCounter;
	//mouseCallback(0);
} END_OF_STATIC_FUNCTION(timerCounterUpdater)

//initialise allegro and some globals
void init()
{
	//initialise Allegro, timer, keyboard and sound
	allegro_init();

	if(install_timer() < 0)
		errors::timerError();
	if(install_keyboard() < 0)
		errors::keyboardError();
	if(install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, NULL) < 0)
		errors::soundDriverError();
	//if(install_mouse() < 0)
	    //errors::mouseError();

	//Set a title for the window
	set_window_title("Shopping in Hell by Master Lenman");

	//load globals and set graphics mode
	global.init();

	clear(global.doublebuffer);
	textout_centre_ex(global.doublebuffer, global.size20, "Loading hell, please wait...", SCREEN_W >> 1, SCREEN_H >> 1, colors::white, -1);
	show();

    //Lock Timer Function and variable
	LOCK_FUNCTION(timerCounterUpdater);
	LOCK_VARIABLE(timerCounter);
	LOCK_FUNCTION(mouseCallback);

    //Set timer to TIMER_FPS fps
	install_int_ex(timerCounterUpdater, BPS_TO_TIMER(TIMER_FPS));

	//Set Charset
	set_uformat(U_ASCII);

	//Initialise rand with the time
	srand(time(0));
	
	/*show_mouse(NULL);
	if(set_display_switch_mode(SWITCH_BACKGROUND) < 0)
	{
		set_display_switch_mode(SWITCH_BACKAMNESIA);
	}*/
}

void show(bool showMouse, BITMAP *bitmap)
{
	if(!bitmap)
	{
		bitmap = global.doublebuffer;
		if(keyStates[KEY_F9])
		{
			save_bitmap((string("media/screenshots/screenshot") + i2str(global.shotnumber++, 5) + string(".bmp")).c_str(), global.doublebuffer, NULL);
		}
	}
	//if(showMouse)
		//drawMouse();
	if(global.vsync)
		vsync();
	blit(bitmap, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
}

BITMAP *loadBitmap(string filename)
{
	BITMAP *retVal = load_bitmap(filename.c_str(), 0);
	if(!retVal)
		errors::loadBitError(filename);
	return retVal;
}

FONT *loadBitmapFont(string filename)
{
	FONT *retVal = load_bitmap_font(filename.c_str(), 0, 0);
	if(!retVal)
	    errors::loadBitError(filename);
	return retVal;
}

BITMAP *createBitmap(int width, int height)
{
	BITMAP *retVal = create_bitmap(width, height);
	if(!retVal)
		errors::createError();
	return retVal;
}

BITMAP *createDatafileBitmap(DATAFILE *datafile, int index)
{
	BITMAP *original = (BITMAP*)datafile[index].dat;
	BITMAP *copy = createBitmap(original->w, original->h);
	blit(original, copy, 0, 0, 0, 0, original->w, original->h);
	return copy;
}

void setMode(int &mode)
{
	if(set_gfx_mode(mode, SCREENW, SCREENH, 0, 0) < 0)
	{
		mode ^= 3;
		if(set_gfx_mode(mode, SCREENW, SCREENH, 0, 0) < 0)
			errors::graphicModeError();
	}
}

void setMode(char &mode)
{
	int temp = mode;
	setMode(temp);
	mode = temp;
}

bool fileSelect(string message, string &path, string ext, int w, int h)
{
	char pathBuffer[1024];
	strncpy(pathBuffer, path.c_str(), 1024);
	int retVal = file_select_ex(message.c_str(), pathBuffer, ext.c_str(), 1024, w, h);
	if(!retVal)
		return false;
	path = pathBuffer;
	return true;
}

bool fileExists(string filename)
{
	ifstream testStream(filename.c_str());
	bool result = !testStream.fail();
	testStream.close();

	return result;
}

string i2str(int i, int minLength)
{
	ostringstream oss; // create a stream
	oss << i; // insert value to stream
	if((minLength -= oss.str().length()) <= 0)
		return oss.str(); // extract value and return it
	string retVal;
	while(retVal.length() < static_cast<unsigned int>(minLength))
		retVal.append(1, '0');
	return retVal + oss.str();
}

string i2str(int i)
{
	ostringstream oss; // create a stream
	oss << i; // insert value to stream
	return oss.str(); // extract value and return it
}

string f2str(float i)
{
	ostringstream oss; // create a stream
	oss << i; // insert value to stream
	return oss.str(); // extract value and return it
}

//Turns a string into an integer, returns defaultValue if the string is no int value
int str2i(const std::string str, const int defaultValue)
{
	istringstream oss(str);
	int r;
	if(!(oss >> r))
		r = defaultValue;
	return r;
}

//Turns a string into an floating point number, returns defaultValue if the string is no
//floating point value
float str2f(const std::string str, const float defaultValue)
{
	istringstream oss(str);
	float r;
	if(!(oss >> r))
		r = defaultValue;
	return r;
}
